
// bibhtml: Generates HTML page (with images) from BibTex files
// See: https://www.telecom-paris.fr/~elc/software/
//
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com)
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef __BibDefs__
#define __BibDefs__

#include <string>
#include "ccstringset.hpp"


/** @brief Entry Types.
 *
 * Entries can be of any type. Regular BibTex types, common biblatex types and
 * special bibtohtml types are predefined (see macro BIBTYPES). Other types found 
 * in BibTex files are automatically added.
 *
 * Specific types:
 *  - _style_: this entry is a printing style (see BibStyle).
 *  - _authgroup_: this entry is a list of authors (see BibGroup).
 *  - _knownmedia_: this entry is a media (a journal, a conference, etc. type does not matter).
 *     * Its _title_ field is the name of the media.
 *     * Its _rank_ field can be used to rank entries corresponding to media
 *       automatically (see _media_ and _rank_ entry fields in BibField).
 */
class BibTypes {
  /// Predefined entry types.
#define BIBTYPES \
none, article, book, booklet, collection, inbook, incollection, inproceedings, \
manual, mastersthesis, misc, phdthesis, proceedings, techreport, unpublished, \
conference, electronic, online, patent, reference, report, www, \
knownmedia, style, authgroup

public:
  /// Predefined type IDs.
  enum ID {BIBTYPES};
  
  /// returns the typeName corresponding to this ID (or an empty string if this ID is unknown).
  static const std::string& typeName(BibTypes::ID typeID);
  
  /// returns the ID corresponding to this typeName (after adding it if neeeded).
  static BibTypes::ID addType(const std::string& typeName);
  
  /// returns the current number of types.
  static int count();
  
private:
  BibTypes() : names(STRINGIFY(BIBTYPES)) {}    // inits type names.
  
  static BibTypes instance; // singleton
  ccuty::istringset names;  // stores type names

#undef BIBTYPES
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/** @brief Entry Fields
 *
 * Entries can contain any field. Regular BibTex fields, common biblatex fields
 * and special bibtohtml fields are predefined (see macro BIBFIELDS). Other fields 
 * found in BibTex files are automatically added.
 *
 * Specific fields or behaviors:
 *  - _abstract_, _keywords_, _award_ : corresponding text. By default the abstract
 *     and the keywords are displayed on demand by clicking on a hyperlink.
 *  - _image_ _video_, _link_, _poster_, _project_, _slides_, _software_: corresponding URLs.
 *  - _bibcite_: BibTex citation (set automatically).
 *  - _bibkey_: BibTex key (set automatically). Not to be confused with the _key_ field.
 *  - _media_: unified "media" of the entry (set automatically). Depending on the
 *     entry type, this field will contain the value of _journal_ (for \@article),
 *     _booktitle_ (for \@inproceedings or similar) or _title_.
 *  - _rank_: entry rank, can serve to sort entries according to their rank.
 *     Values can be "0", "1", "2", etc. or "?" (unknown). This field will be set
 *     automatically if a set of "knownmedia" (see BibTypes) is provided.
 *
 * Depending on the printing style (see BibStyle), fields can be displayed as images
 * keywords, or plain text. They can also be displayed on demand (using javascript).
 */
class BibFields {
  /// Predefined entry fields.
#define BIBFIELDS \
none, address, abstract, annote, author, booktitle, chapter, crossref, \
edition, editor, eprint, howpublished, institution, journal, key, month, note, \
number, organization, pages, publisher, school, series, title, type, url, volume, year, \
doi, introduction, issn, issue, journaltitle, language, location, \
pagetotal, pdf, pubstate, subtitle, venue, \
award, hal, image, keywords, link, media, slides, poster, project, video, software,\
bibkey, bibcite, rank, authgroup
  
public:
  /// Predefined field IDs.
  enum ID {BIBFIELDS};
  
  /// returns the fieldName corresponding to this ID (or an empty string if this ID is unknown).
  static const std::string& fieldName(BibFields::ID fieldID);
  
  /// returns the ID corresponding to this fieldName (after adding it if neeeded).
  static BibFields::ID addField(const std::string& fieldName);
  
  /** returns the ID corresponding to this fieldName.
   *  The name can be abridged if _abridged_ is true.
   *  Returns BibField::none (i.e. 0) if _fieldName_ is unknown.
   */
  static BibFields::ID fieldID(const std::string& fieldName, bool abridged);
  
  
private:
  /// Constructor: initializes predefined entry fields.
  BibFields() : names(STRINGIFY(BIBFIELDS)) {}
  
  static BibFields instance; // singleton
  ccuty::istringset names;   // stores field names

#undef BIBFIELDS
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/** @brief Default settings.
 */
class BibSettings {
public:
  /// Default pathname of the HTML file that is generated.
  static constexpr const char* OutHtml = "out.html";

  /// Default pathname of the BibTex file that is generated.
  static constexpr const char* OutBib = "out.bib";
  
  /// Default pathname of the HTML template file.
  static constexpr const char* TemplateFile = "template.html";

  /// The HTML template file must contain this string (that will be replaced by the entries).
  static constexpr const char* TemplateMark = "###CONTENT";

  /// URL prefix for DOIs.
  static constexpr const char* DoiPrefix = "https://dx.doi.org/";

  /// URL prefix for HAL IDs.
  static constexpr const char* HalPrefix = "https://hal.archives-ouvertes.fr/";
  
  /** @brief Format of the default printing Style.
   * This printing Style is used by default if none is specified (see BibStyle).
   */
  static const char* DefaultStyleFormat;
};

#endif
