
// bibhtml: Generates HTML page (with images) from BibTex files
// See: https://www.telecom-paris.fr/~elc/software/
//
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com)
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "bib.hpp"
#include "bibentry.hpp"
#include "bibstyle.hpp"

#define FIELD_KEY '%'
#define LINK_KEY '^'
#define BUTTON_KEY '?'
#define ICON_KEY '<'
#define ICON_END_KEY '>'


void TypeStyle::parse(Bib& btw, std::string pattern) {
  fields.clear();
  const char * begin = pattern.c_str();
  const char * end = begin + pattern.size();
  const char * p = begin;
  FieldStyle* f = nullptr;

  while (p < end) {
    char enclosed = 0, enclosedEnd = 0;
    f = new FieldStyle();

    if (*p != FIELD_KEY) {    // before or between fields
      const char * p2 = p;
      while (p2 < end && *p2 != FIELD_KEY) ++p2;
      
      f->type = FieldStyle::Separator;
      f->prefix.assign(pattern, p-begin, p2-p);
      size_t pos = f->prefix.find("\n");
      if (pos != std::string::npos) {   // contient \n
        f->breakline = true;
        f->brprefix = f->prefix;
        f->brprefix.replace(pos, 1, "<br>");
        std::remove(f->prefix.begin(), f->prefix.end(), '\n');
      }
      fields.push_back(f);
      p = p2;
    }
    
    else {    // parsing field
      if (p+1 >= end) break;   // ignoring % at the end of the line
      
      else if (*(p+1) == '{' || *(p+1) == '(' || *(p+1) == '[') {
        enclosed = *(p+1);
        if (enclosed == '{') enclosedEnd = '}';
        else if (enclosed == '(') enclosedEnd = ')';
        else if (enclosed == '[') enclosedEnd = ']';
        p += 2;
        const char * p2 = p;
        while (p2 < end && *p2 != FIELD_KEY && *p2 != enclosedEnd) ++p2;
        if (p < end && p2 > p) f->prefix.assign(pattern, p-begin, p2-p);
        p = p2;
        if (p >= end) break;
        else if (*p == enclosedEnd) {fields.push_back(f); continue;}
      }
      
      if (*(p+1) == LINK_KEY) {f->type = FieldStyle::Link; p++;}
      else if (*(p+1) == BUTTON_KEY) {f->type = FieldStyle::Button; p++;}
      else f->type = FieldStyle::Text;

      const char * p2 = ++p;
      while (p2 < end && isalnum(*p2)) p2++;
      f->name.assign(pattern, p-begin, p2-p);
      f->field = BibFields::addField(f->name);  // add field to BibFields if needed
      fields.push_back(f);
      
      if (*p2 == ICON_KEY) {   // icon
        p = p2;
        while (p2 < end && *p2 != ICON_END_KEY) p2++;
        if (*p2 == ICON_END_KEY) {
          f->img.assign(pattern, p-begin+1, p2-p-1);
          ccuty::strtrim(f->img);
          p2++;
        }
        else btw.echo("Error in style: A Field contains a < without an ending > in: "+pattern);
      }
      
      p = p2;
      if (enclosed && p < end) {
        if (*p == enclosedEnd) p++;
        else {
          const char * p2 = p;
          while (p2 < end && *p2 != enclosedEnd) p2++;
          if (p2 > p) f->suffix.assign(pattern, p-begin, p2-p);
          p = p2+1;
        }
      }
    }
  }
  
  if (f && f->type == FieldStyle::None) delete f;
}


bool TypeStyle::write(Bib& btw, std::ostream& f, const BibEntry& entry, int number) const {
  if (fields.empty()) return false;
  
  std::string bibkey = entry.field(BibFields::bibkey);
  if (bibkey.empty()) {
    btw.echo("Error in BibText file: Entry "+ entry.bibkey() + "has no BibTex key: ");
    return false;
  }
  
  // create a "cleaned" version of the bibtex key generated by replacing all : with -.
  replace(bibkey.begin(), bibkey.end(), ':', '-');
  bool previous = false;

  f << "<tr class=\"entry\">" << std::endl;

  // number
  if (btw.showNumbers()) f << "<td class=\"number\">[" << number << "]</td>" << std::endl;
  
  if (btw.showImages() == ShowImages::Left)  {
    f << "<td class=\"image\">";
    const std::string& ima = entry.field(BibFields::image);
    if (ima.empty()) f << "&nbsp;";
    else {
      f << "<img class=\"image\" src=\""<< btw.imageDir() << ima << "\"/>";
    }
    f << "</td>\n";
  }
  
  // body
  f << "<td class=\"ebody\">" << std::endl;
  f << "<a name=\"" << bibkey << "\"></a>"; // anchor for the bibtex citation
  
  std::string separ;
  for (auto it : fields) {
    
    if (!it) continue;

    if (it->type == FieldStyle::Separator) {
      if (previous && !it->prefix.empty()) separ = it->prefix;
      if (it->breakline) {
        if (btw.breakLines()) separ = it->brprefix; else separ = it->prefix;
      }
      previous = false;
      continue;
    }
    
    const std::string& value = entry.field(it->field);
    if (value.empty()) continue;
    
    switch (it->type) {
      case FieldStyle::Text:
        if (!separ.empty()) f << separ;
        f << '\n';
        if (!it->prefix.empty()) f << it->prefix;
    
        f << "<span class=\""<< it->name <<"\">"<< value <<"</span>";
        
        if (!it->suffix.empty()) f << it->suffix;
        previous = true;
        break;
        
      case FieldStyle::Link:
        if (btw.showLinks() != ShowLinks::None) {
          if (!separ.empty()) f << separ;
          f << '\n';
          if (!it->prefix.empty()) f << it->prefix;
          
          f << "<a class=\""<< it->name <<"\" href=\""<< value <<"\" target=\"bis\">";
          
          if (btw.showLinks()== ShowLinks::Text || it->img.empty())
            f << it->name <<"</a>";
          else
            f << "<img class=\"" << it->name <<"\" alt=\""<< it->name
            <<"\" src=\"" << btw.imageDir() << it->img << "\"/></a>";
          
          if (!it->suffix.empty()) f << it->suffix;
          previous = true;
        }
        break;
        
      case FieldStyle::Button:
        if (btw.showLinks() != ShowLinks::None) {
          if (!separ.empty()) f << separ;
          f << '\n';
          if (!it->prefix.empty()) f << it->prefix;
          const std::string& key = entry.field(BibFields::bibkey);
      
          // button link
          f << "<a class=\""<< it->name
          <<"\" href=\"javascript:void(0)\" onclick=\"show('"<<key<<"/"<< it->name<<"')\">";
          
          if (btw.showLinks()== ShowLinks::Text || it->img.empty())
            f << it->name <<"</a>\n";
          else
            f << "<img class=\"" << it->name <<"\" alt=\""<< it->name
            <<"\" src=\"" << btw.imageDir() << it->img << "\"/></a>\n";

          // button content
          f << "<div class=\""<< it->name<<"-content\" id=\""<<key<<"/"<< it->name <<"\">"
          << value << "</div>";
          
          if (!it->suffix.empty()) f << it->suffix;
          previous = true;
        }
        break;
        
      default:
        break;
    }
  }
  
  f << "<br>\n";  // final br
  if (btw.verboseMode()) {
    f << "Rank: "<< entry.field(BibFields::rank)
    << "Media: "<< entry.field(BibFields::media) << "<br>\n";
  }
  f << "</td>" << std::endl;

  if (btw.showImages() == ShowImages::Right)  {
    f << "<td class=\"image\">";
    const std::string& image = entry.field(BibFields::image);
    if (image.empty()) f << "&nbsp;";
    else f << "<img class=\"image\" src=\""<< btw.imageDir() << image << "\"/>";
    f << "</td>\n";
  }
  
  f << "</tr>\n" << std::endl;
  return true;
}

/*
 void RefStyle::printFormat() const {
   for (auto& f : fields) {
     cout << "["<<f.prefix<<"]["<<f.fieldName<<"]"<< "["<<f.postfix<<"] type:"<< f.type << endl;
   }
 }
*/
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

BibStyle* BibStyle::createStyle(Bib& btw, const std::string& format) {
  BibStyle* s = new BibStyle();
  s->defaultTypeStyle_.parse(btw, format);
  return s;
}

BibStyle* BibStyle::createStyle(Bib& btw, const BibEntry& styleRef) {
  std::string format = styleRef.field(BibFields::media);
  if (format.empty()) {
    btw.echo("Error in style: The 'media' field should be defined in style " + styleRef.bibkey());
    return nullptr;
  }

  BibStyle* s = new BibStyle();
  s->defaultTypeStyle_.parse(btw, format);
  
  for (auto& it : styleRef.fields()) {
    if (it.first != BibFields::type && it.first != BibFields::bibkey
        && it.first != BibFields::media && it.first != BibFields::rank) {

    // convert the BibField of the ref (e.g. article) into a BibType
    BibTypes::ID type = BibTypes::addType(BibFields::fieldName(it.first));
    auto pair = s->typeStyles_.emplace(type, TypeStyle());
    
      if (pair.second) {  // actually added to _typeStyles
        pair.first->second.parse(btw, it.second.parsed);
      }
      else  // was already present in _typeStyles
        Bib::echo("Error in style: The field " + BibFields::fieldName(it.first)
                  + " is redefined in style " + styleRef.bibkey());
    }
  }
  return s;
}

void BibStyle::write(Bib& btw, const std::string& filename, const BibEntries& refs) const {
  std::ofstream f(filename);
  if (!f) {
    btw.echo("Can't open '"+filename+"' for producing HTML file");
    return;
  }
  f << btw.htmlBegin() << std::endl;
  f << "<table>" << std::endl;
  
  int number = 0;
  for (auto& it : refs) {
    if (!it.isAuxiliary()) number++;
  }
  
  std::string lastHeader;
  for (auto& it : refs) {
    if (it.isAuxiliary()) continue;
    if (btw.showHeaders() != BibFields::none) {
      const std::string& header = it.field(btw.showHeaders());
      if (header != lastHeader) {
        f << "<tr class=\"header\"><td class=\"header\">" << header << "</td></tr>" << std::endl;
        lastHeader = header;
      }
    }
    auto st = typeStyles_.find(it.getTypeID());
    if (st != typeStyles_.end()) st->second.write(btw, f, it.entry_, number);
    else defaultTypeStyle_.write(btw, f, it.entry_, number);
    number--;
  }
  
  f << "</table>" << std::endl;
  f << btw.htmlEnd() << std::endl;
  f.close();
}

